/*-
 * Copyright (c) 2011-2012       Scott Ringwelski <sgringwe@mtu.edu>
 *
 * Originally Written by Scott Ringwelski for BeatBox Music Player
 * BeatBox Music Player: http://www.launchpad.net/beat-box
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

using Gtk;
using Gee;
using Notify;

public class BeatBox.LibraryWindow : LibraryWindowInterface, Gtk.Window {

	// signals
	public signal void playPauseChanged ();

	public static Granite.Application app { get; private set; }
	public BeatBox.LibraryManager lm { get; private set; }
	public BeatBox.Settings settings { get; private set; }
	private BeatBox.MediaKeyListener mkl;
	public BeatBox.SearchSuggester search_suggester;

	private bool queriedlastfm; // whether or not we have queried last fm for the current media info
	private bool media_considered_played; // whether or not we have updated last played and added to already played list
	private bool added_to_play_count; // whether or not we have added one to play count on playing media
	private bool tested_for_video; // whether or not we have tested if media is video and shown video
	private bool scrobbled_track;

	public bool dragging_from_music { get; set; }

	public bool initializationFinished { get; private set; }

	private VBox verticalBox;
	public Notebook mainViews { get; private set; }
	public DrawingArea videoArea  { get; private set; }
	public HPaned sourcesToMedias { get; private set; } //allows for draggable
	public HPaned mediasToInfo { get; private set; } // media info pane
	private ScrolledWindow sideTreeScroll;
	private VBox contentBox;
	public SideTreeView sideTree { get; private set; }
	private InfoPanel infoPanel;
	
	private InfoBar infoBar;
	private Label infoBarLabel;
	LinkedList<Media> file_not_found_pileup;
	
	private Toolbar topControls;
	private ToolButton previousButton;
	private ToolButton playButton;
	private ToolButton nextButton;
	public Granite.Widgets.ModeButton viewSelector;
	public TopDisplay topDisplay { get; private set; }
	public Granite.Widgets.SearchBar  searchField  { get; private set; }
	private StatusBar statusBar;

	private SimpleOptionChooser addPlaylistChooser;
	Gtk.Menu newPlaylistMenu;
	Gtk.ImageMenuItem newPlaylistItem;
	Gtk.ImageMenuItem newSmartPlaylistItem;
	public SimpleOptionChooser shuffleChooser;
	private SimpleOptionChooser repeatChooser;
	private SimpleOptionChooser infoPanelChooser;

	// basic file stuff
	private Gtk.Menu settingsMenu;
	private Gtk.Menu libraryOperationsMenu;
	private ImageMenuItem libraryOperations;
	private Gtk.MenuItem fileSetMusicFolder;
	private Gtk.MenuItem fileImportMusic;
	private Gtk.MenuItem fileRescanMusicFolder;
	private Gtk.MenuItem editEqualizer;
	private ImageMenuItem editPreferences;

	public Notify.Notification notification { get; private set; }

	public LibraryWindow(Granite.Application app, BeatBox.Settings settings, string[] args) {
		this.app = app;
		this.settings = settings;

		// Init LibNotify
		Notify.init ("beatbox");

		// Load icon information
		Icons.init ();

		//this is used by many objects, is the media backend
		lm = new BeatBox.LibraryManager(settings, this, args);
		
		//various objects
		mkl = new MediaKeyListener(lm, this);

#if HAVE_INDICATE
#if HAVE_DBUSMENU
		message("Initializing MPRIS and sound menu\n");
		var mpris = new BeatBox.MPRIS(lm, this);
		mpris.initialize();
#endif
#endif

		dragging_from_music = false;

		this.lm.player.end_of_stream.connect(end_of_stream);
		this.lm.player.current_position_update.connect(current_position_update);
		//FIXME? this.lm.player.media_not_found.connect(media_not_found);
		this.lm.music_counted.connect(musicCounted);
		this.lm.music_added.connect(musicAdded);
		this.lm.music_imported.connect(musicImported);
		this.lm.music_rescanned.connect(musicRescanned);
		this.lm.progress_notification.connect(progressNotification);
		this.lm.medias_updated.connect(medias_updated);
		this.lm.medias_removed.connect(medias_removed);
		this.lm.media_played.connect(media_played);
		this.lm.playback_stopped.connect(playback_stopped);

		this.destroy.connect (on_quit);

		if(lm.media_count() == 0 && settings.getMusicFolder() == "") {
			message("First run.\n");
		}
		else {
			lm.clearCurrent();
			
			// make sure we don't re-count stats
			if((int)settings.getLastMediaPosition() > 5)
				queriedlastfm = true;
			if((int)settings.getLastMediaPosition() > 30)
				media_considered_played = true;
			if(lm.media_active && (double)((int)settings.getLastMediaPosition()/(double)lm.media_info.media.length) > 0.90)
				added_to_play_count = true;

			// rescan on startup
			/*lm.rescan_music_folder();*/
		}
	}

	public void build_ui() {
		// simple message to terminal
		message ("Building user interface\n");
		
		// set the size based on saved gconf settings
		set_default_size(settings.getWindowWidth(), settings.getWindowHeight());
		resize(settings.getWindowWidth(), settings.getWindowHeight());

		// set window min/max
		Gdk.Geometry geo = Gdk.Geometry();
		geo.min_width = 700;
		geo.min_height = 400;
		set_geometry_hints(this, geo, Gdk.WindowHints.MIN_SIZE);

		// set the title
		set_title("BeatBox");

		// set the icon
		set_icon(Icons.BEATBOX.render (IconSize.MENU, null));

		/* Initialize all components */
		verticalBox = new VBox(false, 0);
		sourcesToMedias = new HPaned();
		mediasToInfo = new HPaned();
		contentBox = new VBox(false, 0);
		mainViews = new Notebook ();
		videoArea = new DrawingArea ();

		sideTree = new SideTreeView(lm, this);
		sideTreeScroll = new ScrolledWindow(null, null);
		libraryOperations = new ImageMenuItem.from_stock("library-music", null);
		libraryOperationsMenu = new Gtk.Menu();
		fileSetMusicFolder = new Gtk.MenuItem.with_label(_("Set Music Folder"));
		fileImportMusic = new Gtk.MenuItem.with_label(_("Import to Library"));
		fileRescanMusicFolder = new Gtk.MenuItem.with_label(_("Rescan Music Folder"));
		editEqualizer = new Gtk.MenuItem.with_label(_("Equalizer"));
		editPreferences = new ImageMenuItem.from_stock(Gtk.Stock.PREFERENCES, null);
		settingsMenu = new Gtk.Menu();
		topControls = new Toolbar();
		previousButton = new ToolButton.from_stock(Gtk.Stock.MEDIA_PREVIOUS);
		playButton = new ToolButton.from_stock(Gtk.Stock.MEDIA_PLAY);
		nextButton = new ToolButton.from_stock(Gtk.Stock.MEDIA_NEXT);
		topDisplay = new TopDisplay(lm);
		viewSelector = new Granite.Widgets.ModeButton();
		searchField = new Granite.Widgets.SearchBar(_("Search..."));
		search_suggester = new SearchSuggester(lm, this);
		infoBarLabel = new Label("");
		infoBar = new InfoBar();
		file_not_found_pileup = new LinkedList<Media>();

		infoPanel = new InfoPanel(lm, this);
		statusBar = new StatusBar();

		var add_playlist_image = Icons.render_image ("list-add-symbolic", IconSize.MENU);
		var shuffle_on_image = Icons.SHUFFLE_ON.render_image (IconSize.MENU);
		var shuffle_off_image = Icons.SHUFFLE_OFF.render_image (IconSize.MENU);
		var repeat_on_image = Icons.REPEAT_ON.render_image (IconSize.MENU);
		var repeat_off_image = Icons.REPEAT_OFF.render_image (IconSize.MENU);
		var info_panel_show = Icons.PANE_SHOW_SYMBOLIC.render_image (IconSize.MENU);
		var info_panel_hide = Icons.PANE_HIDE_SYMBOLIC.render_image (IconSize.MENU);

		addPlaylistChooser = new SimpleOptionChooser.from_image (add_playlist_image);
		shuffleChooser = new SimpleOptionChooser.from_image (shuffle_on_image, shuffle_off_image);
		repeatChooser = new SimpleOptionChooser.from_image (repeat_on_image, repeat_off_image);
		infoPanelChooser = new SimpleOptionChooser.from_image (info_panel_hide, info_panel_show);

		infoPanelChooser.setTooltip (_("Hide Info Panel"), _("Show Info Panel"));
		shuffleChooser.setTooltip (_("Turn off Shuffle"), _("Turn on Shuffle"));
		repeatChooser.setTooltip (_("Turn off Repeat"), _("Turn on Repeat"));
		addPlaylistChooser.setTooltip (_("Add Playlist"));
		
		// Menu to let user right click add to playlist and choose smart playlist instead
		newPlaylistMenu = new Gtk.Menu();
		newPlaylistItem = new Gtk.ImageMenuItem.with_label("Create new Playlist");
		newSmartPlaylistItem = new Gtk.ImageMenuItem.with_label("Create new Smart Playlist");
		newPlaylistItem.set_image(new Image.from_gicon(Icons.PLAYLIST.get_gicon(), Gtk.IconSize.MENU));
		newSmartPlaylistItem.set_image(new Image.from_gicon(Icons.SMART_PLAYLIST.get_gicon(), Gtk.IconSize.MENU));
		newPlaylistMenu.append(newPlaylistItem);
		newPlaylistMenu.append(newSmartPlaylistItem);

		statusBar.insert_widget (addPlaylistChooser as Gtk.Widget, true);
		statusBar.insert_widget (new Gtk.Box (Orientation.HORIZONTAL, 12), true);
		statusBar.insert_widget (shuffleChooser as Gtk.Widget, true);
		statusBar.insert_widget (repeatChooser as Gtk.Widget, true);
		statusBar.insert_widget (infoPanelChooser as Gtk.Widget);

		notification = new Notify.Notification ("", null, null);

		/* Set properties of various controls */
		sourcesToMedias.set_position(settings.getSidebarWidth());
		mediasToInfo.set_position((lm.settings.getWindowWidth() - lm.settings.getSidebarWidth()) - lm.settings.getMoreWidth());

		/* ADD MAIN VIEWS */
		build_main_views ();
	
		/* ADD PLAYLIST VIEWS */
		load_playlists ();

		/* LOAD MUSIC STORE VIEW */
		load_default_store ();
		message("all done\n");

		sideTreeScroll = new ScrolledWindow(null, null);
		//FIXME: don't scroll horizontally
		sideTreeScroll.set_policy (PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
		sideTreeScroll.add(sideTree);

		update_sensitivities();

		/* create appmenu menu */
		editPreferences.set_label(_("Preferences"));
		
		settingsMenu.append(fileImportMusic);
		settingsMenu.append(fileRescanMusicFolder);
		settingsMenu.append(new SeparatorMenuItem());
		settingsMenu.append(editEqualizer);
		settingsMenu.append(editPreferences);

		fileSetMusicFolder.activate.connect(editPreferencesClick);
		fileImportMusic.activate.connect(fileImportMusicClick);
		fileRescanMusicFolder.activate.connect(fileRescanMusicFolderClick);
		editEqualizer.activate.connect(editEqualizerClick);
		editPreferences.activate.connect(editPreferencesClick);

		repeatChooser.appendItem(_("Off"));
		repeatChooser.appendItem(_("Song"));
		repeatChooser.appendItem(_("Album"));
		repeatChooser.appendItem(_("Artist"));
		repeatChooser.appendItem(_("All"));

		shuffleChooser.appendItem(_("Off"));
		shuffleChooser.appendItem(_("All"));

		infoPanelChooser.appendItem(_("Hide"));
		infoPanelChooser.appendItem(_("Show"));

		/* Add controls to the GUI */
		add(verticalBox);
		verticalBox.pack_start(topControls, false, true, 0);
		verticalBox.pack_start(videoArea, true, true, 0);
		verticalBox.pack_start(sourcesToMedias, true, true, 0);
		verticalBox.pack_end(statusBar, false, true, 0);

		ToolItem topDisplayBin = new ToolItem();
		ToolItem viewSelectorBin = new ToolItem();
		ToolItem searchFieldBin = new ToolItem();
		
		var viewSelectorContainer = new Box (Orientation.VERTICAL, 0);
		var viewSelectorInnerContainer = new Box (Orientation.HORIZONTAL, 0);
		viewSelectorInnerContainer.pack_start (new Box (Orientation.HORIZONTAL, 10), true, true, 0);
		viewSelectorInnerContainer.pack_start (viewSelector, false, false, 0); // VIEW SELECTOR
		viewSelectorInnerContainer.pack_end (new Box (Orientation.HORIZONTAL, 10), true, true, 0);
		viewSelectorContainer.pack_start (new Box (Orientation.VERTICAL, 5), true, true, 0);
		viewSelectorContainer.pack_start (viewSelectorInnerContainer, false, false, 0);
		viewSelectorContainer.pack_end (new Box (Orientation.VERTICAL, 5), true, true, 0);
		
		topDisplayBin.add(topDisplay);
		topDisplayBin.set_expand(true);
		
		viewSelectorBin.add(viewSelectorContainer);
		viewSelectorBin.margin_left = 12;
		viewSelectorBin.margin_right = 12;
		
		searchFieldBin.add(searchField);
		searchFieldBin.margin_left = 12;
		searchFieldBin.margin_right = 12;
		
		// Info bar to give notices about file not found errors
		infoBar.set_message_type(Gtk.MessageType.ERROR);
		var action_hbox = new HBox(false, 0);
		((Gtk.Container)infoBar.get_action_area()).add(action_hbox);
		
		var resolve_button = new Button.with_label("Resolve");
		//var ignore_button = new Button.with_label("Ignore");
		action_hbox.pack_start(resolve_button, false, true, 0);
		//action_hbox.pack_start(ignore_button, false, true, 6);
		
		resolve_button.clicked.connect( () => { info_bar_resonse(Gtk.ResponseType.YES); });
		//ignore_button.clicked.connect( () => { info_bar_resonse(Gtk.ResponseType.NO); });
		
		infoBarLabel.set_justify(Justification.LEFT);
		infoBarLabel.set_single_line_mode(true);
		infoBarLabel.ellipsize = Pango.EllipsizeMode.END;
		
		((Gtk.Container)infoBar.get_content_area()).add(infoBarLabel);

		// Set theming
		topControls.get_style_context().add_class(STYLE_CLASS_PRIMARY_TOOLBAR);
		sourcesToMedias.get_style_context().add_class ("sidebar-pane-separator");

		topControls.set_vexpand (false);
		topControls.set_hexpand (true);
		
		viewSelector.append(Icons.VIEW_DETAILS.render_image (IconSize.MENU));
		viewSelector.append(Icons.VIEW_ICONS.render_image (IconSize.MENU));

		topControls.insert(previousButton, -1);
		topControls.insert(playButton, -1);
		topControls.insert(nextButton, -1);
		topControls.insert(viewSelectorBin, -1);
		topControls.insert(topDisplayBin, -1);
		topControls.insert(searchFieldBin, -1);
		topControls.insert(app.create_appmenu(settingsMenu), -1);

		// Hide notebook tabs and border
		mainViews.show_tabs = false;
		mainViews.show_border = false;
		
		contentBox.pack_start(infoBar, false, true, 0);
		contentBox.pack_start(mainViews, true, true, 0);

		mediasToInfo.pack1(contentBox, true, true);
		mediasToInfo.pack2(infoPanel, false, false);

		sourcesToMedias.pack1(sideTreeScroll, false, true);
		sourcesToMedias.pack2(mediasToInfo, true, true);

		// add mounts to side tree view
		lm.device_manager.loadPreExistingMounts();
		
		viewSelector.selected = 0;// Always use 0 to give grid time to situate itselfsettings.getViewMode();
		
		int i = settings.getLastMediaPlaying();
		if(i != 0 && lm.media_from_id(i) != null && File.new_for_uri(lm.media_from_id(i).uri).query_exists()) {
			lm.playMedia(lm.media_from_id(i), true);
			
			if(lm.media_from_id(i).resume_pos > (int)((double)lm.media_from_id(i).length * 0.90f)) // don't add to play count twice
				added_to_play_count = true;
		}
		else {
			// don't show info panel if nothing playing
			infoPanel.set_visible(false);
		}

		/* Connect events to functions */
		previousButton.clicked.connect(previousClicked);
		playButton.clicked.connect(playClicked);
		nextButton.clicked.connect(nextClicked);
		viewSelector.mode_changed.connect( () => { 
			settings.setViewMode(viewSelector.selected);
			update_sensitivities();
			//get_current_view_wrapper().set_as_current_view();
		 } );
		addPlaylistChooser.button_press_event.connect(addPlaylistChooserOptionClicked);
		newPlaylistItem.activate.connect(sideTree.playlistMenuNewClicked); // same as above
		newSmartPlaylistItem.activate.connect(sideTree.smartPlaylistMenuNewClicked);
		repeatChooser.option_changed.connect(repeatChooserOptionChanged);
		shuffleChooser.option_changed.connect(shuffleChooserOptionChanged);
		infoPanelChooser.option_changed.connect(infoPanelChooserOptionChanged);
		searchField.activate.connect(searchFieldActivate);
		key_press_event.connect(keyPressed);
		infoBar.response.connect(info_bar_resonse);

		/* set up drag dest stuff */
		drag_dest_set(this, DestDefaults.ALL, {}, Gdk.DragAction.MOVE);
		drag_dest_add_uri_targets(this);
		drag_data_received.connect(dragReceived);
		initializationFinished = true;

		sideTree.resetView();

		var vw = (ViewWrapper)get_current_view_wrapper ();
		if(lm.media_active)
			vw.list_view.set_as_current_list();
		
		lm.set_shuffle_mode((LibraryManager.Shuffle)settings.getShuffleMode());
		lm.set_repeat_mode((LibraryManager.Repeat)settings.getRepeatMode());
		repeatChooser.setOption(settings.getRepeatMode());
		shuffleChooser.setOption(settings.getShuffleMode());
		infoPanelChooser.setOption(settings.getMoreVisible() ? 1 : 0);

		resize(settings.getWindowWidth(), this.default_height);
		show_all();
		
		// Hide some things
		infoBar.hide();
		videoArea.hide();
		
		searchField.set_text(lm.settings.getSearchString());

		update_sensitivities();
		
		// If we have no songs in library, try and get some from default music folder
		if(lm.song_ids().size == 0)
			setMusicFolder(Environment.get_user_special_dir(UserDirectory.MUSIC));
			
		// Enable video now that drawingArea is set up
		//lm.player.setVideoEnabled(true);
	}

	public static Gtk.Alignment wrap_alignment (Gtk.Widget widget, int top, int right, int bottom, int left) {
		var alignment = new Gtk.Alignment(0.0f, 0.0f, 1.0f, 1.0f);
		alignment.top_padding = top;
		alignment.right_padding = right;
		alignment.bottom_padding = bottom;
		alignment.left_padding = left;

		alignment.add(widget);
		return alignment;
	}


	/**
	 * Description:
	 * Builds the views (view wrapper) and adds the respective element to the sidebar TreeView.
	 *
	 * @param tree The sidebar tree to build it on [if NULL is passed it uses the default tree]
	 * @param hint The type of View Wrapper
	 * @param view_name The name of the item in the sidebar
	 * @param media indexes of the media to show in the view
	 * @param sort_column treeview column used to sort list items
	 * @param sort_column column used to sort the list. Default: "" [allow-none]
	 * @param sort_type sort type (ascending, descending, etc.) Default: 0 [allow-none]
	 */
	public TreeIter add_view (ViewWrapper.Hint hint, string view_name,
	                              Collection<Media> media, TreeViewSetup tvs, 
	                              int id = -1)
	{
		TreeIter rv;
		ViewWrapper view_wrapper;
		
		if(hint == ViewWrapper.Hint.MUSIC)
			view_wrapper = new MusicViewWrapper(this, media, tvs, id);
		else if(hint == ViewWrapper.Hint.PODCAST)
			view_wrapper = new PodcastViewWrapper(this, media, tvs, id);
		else if(hint == ViewWrapper.Hint.STATION)
			view_wrapper = new StationViewWrapper(this, media, tvs, id);
		else if(hint == ViewWrapper.Hint.SIMILAR)
			view_wrapper = new SimilarViewWrapper(this, media, tvs, id);
		else
			view_wrapper = new ViewWrapper (this, media, tvs, id);
		
		GLib.Object? o = null;
		if(hint == ViewWrapper.Hint.PLAYLIST)
			o = lm.playlist_from_id(id);
		else if(hint == ViewWrapper.Hint.SMART_PLAYLIST)
			o = lm.smart_playlist_from_id(id);
		
		rv = sideTree.add_item (view_wrapper, view_name, o);

		/* Pack view wrapper into the main views */
		if (add_to_main_views(view_wrapper) == -1)
			critical ("Failed to append view '%s' to BeatBox's main views", view_name);

		return rv;
	}

	/**
	 * Sets the given view as the active item
	 */
	public void set_active_view (Gtk.Widget view) {
		int view_index = mainViews.page_num (view);
		
		if (view_index < 0) {
			critical ("Cannot set " + view.name + " as the active view");
			return;
		}

		// GtkNotebooks don't show hidden widgets. Make sure we show the view just in case ...
		view.show_all ();

		// We need to set this view as the current page before even attempting to call
		// the set_as_current_view() method.
		mainViews.set_current_page (view_index);
		
		// TODO: abstract this in MainView
		if (view is ViewWrapper) {
			if(!initializationFinished)
				return;

			((ViewWrapper)view).set_as_current_view();
		}
		else if(view is Store.StoreView) {
			((Store.StoreView)view).setIsCurrentView(true);
		}
		else if(view is DeviceView) {
			DeviceView dv = (DeviceView)view;
			dv.set_as_current_view ();
		}
	}

	/**
	 * Appends a widget to the main views.
	 *
	 * WARNING: Don't use this method directly.
	 */
	public int add_to_main_views (Gtk.Widget view) {
		return mainViews.append_page (view);
	}

	/**
	 * Description:
	 * Builds the views (view wrapper) and adds the respective element to the sidebar TreeView.
	 *
	 * @param name The name of the item in the sidebar
	 * @param widget Widget containing the custom view
	 * @param tree The sidebar tree to build it on [if NULL is passed it uses the default tree]
	 *
	 * TODO: add plugin hook and update LibraryWindowInterface.
	 * TODO: Add option to pass an icon (needed by plugins).
	 *
	 * IMPORTANT: Currently every item added through this method will be put under the Network category
	 */
	public ViewWrapper add_custom_view (string name, Gtk.Widget widget, GLib.Object? o) {
		var view_wrapper = new ViewWrapper.with_view (widget);

		sideTree.add_item (view_wrapper, name, o);

		/* Pack view wrapper into the main views */
		int view_index = add_to_main_views (view_wrapper);

		if (view_index == -1)
			critical ("Failed to append view '%s' to BeatBox's main views", name);

		return view_wrapper;
	}

	/**
	 * Builds and sets up the default BeatBox views. That includes main sidebar elements
	 * and categories, which at the same time wrap treeviews, icon views, welcome screens, etc.
	 */
	private void build_main_views () {
		message ("Building main views ...");

		// Add Music Library View
		add_view (ViewWrapper.Hint.MUSIC, _("Music"), lm.songs (), lm.music_setup);

		// Add Podcast Library View
		add_view (ViewWrapper.Hint.PODCAST, _("Podcasts"), lm.podcasts (), lm.podcast_setup);

		// Add Internet Radio View
		add_view (ViewWrapper.Hint.STATION, _("Internet Radio"), lm.stations(), lm.station_setup);

		// Add Similar playlist. FIXME: This is part of LastFM and shouldn't belong to the core in the future
		add_view (ViewWrapper.Hint.SIMILAR, _("Similar"), new LinkedList<Media>(), lm.similar_setup);

		// Add Queue view
		add_view (ViewWrapper.Hint.QUEUE, _("Queue"), lm.queue (), lm.queue_setup);

		// Add History view
		add_view (ViewWrapper.Hint.HISTORY, _("History"), lm.already_played (), lm.history_setup);

		message ("Done with main views.");
	}

	private void load_playlists () {
		debug ("Loading playlists");
		
		// load smart playlists
		foreach(SmartPlaylist p in lm.smart_playlists()) {
			addSideListItem(p);
		}

		// load playlists
		foreach(Playlist p in lm.playlists()) {
			addSideListItem(p);
		}

		debug ("Finished loading playlists");
	}

	private void load_default_store () {
		if (Option.enable_store) {
			var storeView = new Store.StoreView(lm, this);
			add_custom_view (_("Music Store"), storeView, null);
		}
	}

	public TreeIter? addSideListItem(GLib.Object o) {
		TreeIter item = sideTree.library_music_iter; //just a default
		ViewWrapper vw = null;

		if(o is Playlist) {
			Playlist p = (Playlist)o;

			item = add_view (ViewWrapper.Hint.PLAYLIST, p.name, lm.medias_from_playlist(p.rowid),
			           p.tvs, p.rowid);
		}
		else if(o is SmartPlaylist) {
			SmartPlaylist p = (SmartPlaylist)o;
			
			item = add_view (ViewWrapper.Hint.SMART_PLAYLIST, p.name, lm.medias_from_smart_playlist(p.rowid),
			          p.tvs, p.rowid);
		}
		/* XXX: Migrate this code to the new API */
		else if(o is Device) {
			Device d = (Device)o;

			if(d.getContentType() == "cdrom") {
				message("CD added with %d songs.\n", d.get_medias().size);
				vw = new DeviceViewWrapper(this, d.get_medias(), new TreeViewSetup(MusicList.MusicColumn.TRACK, Gtk.SortType.ASCENDING, ViewWrapper.Hint.CDROM), -1, d);
				item = sideTree.addSideItem(sideTree.devices_iter, d, vw, d.getDisplayName(), ViewWrapper.Hint.CDROM);
				add_to_main_views (vw);
			}
			else {
				debug ("adding ipod device view with %d\n", d.get_medias().size);
				DeviceView dv = new DeviceView(lm, d);
				//vw = new DeviceViewWrapper(this, d.get_medias(), "Artist", Gtk.SortType.ASCENDING, ViewWrapper.Hint.DEVICE, -1, d);
				item = sideTree.addSideItem(sideTree.devices_iter, d, dv, d.getDisplayName(), ViewWrapper.Hint.NONE);
				add_to_main_views (dv);
			}
		}
		
		return item;
	}

	/**
	 * This is handled more carefully inside each ViewWrapper object.
	 */
	public void update_sensitivities() {
		if(!initializationFinished)
			return;

		bool folderSet = (lm.settings.getMusicFolder() != "");
		bool haveMedias = lm.media_count() > 0;
		bool doingOps = lm.doing_file_operations();
		bool mediaActive = lm.media_active;
		bool showMore = lm.settings.getMoreVisible();

		bool showingMediaList = (get_current_view_wrapper () is ViewWrapper);
		bool songsInList = showingMediaList ? (get_current_view_wrapper () as ViewWrapper).have_media : false;

		fileSetMusicFolder.set_sensitive(!doingOps);
		fileImportMusic.set_sensitive(!doingOps && folderSet);
		fileRescanMusicFolder.set_sensitive(!doingOps && folderSet);

		if(doingOps)
			topDisplay.show_progressbar();
		else if(mediaActive && lm.media_info.media.mediatype == Media.MediaType.STATION) {
			topDisplay.hide_scale_and_progressbar();
		}
		else
			topDisplay.show_scale();

		topDisplay.set_visible(mediaActive || doingOps);
		topDisplay.set_scale_sensitivity(mediaActive);

		previousButton.set_sensitive(mediaActive || songsInList);
		playButton.set_sensitive(mediaActive || songsInList);
		nextButton.set_sensitive(mediaActive || songsInList);
		
		infoPanel.set_visible(showMore && mediaActive);
		infoPanelChooser.set_visible(mediaActive);

		// hide playlists when media list is empty
		sideTree.setVisibility(sideTree.playlists_iter, haveMedias);
		
		videoArea.set_visible(viewSelector.selected == 2);
		sourcesToMedias.set_visible(viewSelector.selected != 2);

		if(!lm.media_active || haveMedias && !lm.playing) {
			playButton.set_stock_id(Gtk.Stock.MEDIA_PLAY);
		}
	}
	
	bool keyPressed(Gdk.EventKey event) {
		if(event.state != 0) {
			if((event.state & Gdk.ModifierType.CONTROL_MASK) == Gdk.ModifierType.CONTROL_MASK && event.keyval == Gdk.Key.q) {
				message("User ctrl-q quit.\n");
				destroy();
			}
		}
		else if(Regex.match_simple("[a-zA-Z0-9]", event.str) && searchField.sensitive && !searchField.has_focus) {
			if(!(get_focus() is Gtk.Entry)) {
				searchField.grab_focus();
				message("Search field focused\n");
				//lw.searchField.insert_at_cursor(event.str);
			}
		}
		else if(event.str == " " && lm.media_active && !searchField.has_focus) {
			playClicked ();
		}

		return false;
	}

	public virtual void progressNotification(string? message, double progress) {
		if(message != null && progress >= 0.0 && progress <= 1.0)
			topDisplay.set_label_markup(message);

		topDisplay.set_progress_value(progress);
	}

	public void updateInfoLabel() {
		if(lm.doing_file_operations()) {
			debug ("doing file operations, returning null in updateInfoLabel\n");
			
			var label = "";
			if(lm.fo.import_type == FileOperator.ImportType.SET)
				label = "Importing music from <b>" + lm.settings.getMusicFolder() + "</b>...";
			else if(lm.fo.import_type == FileOperator.ImportType.RESCAN)
				label = "Rescanning music folder for changes...";
			else if(lm.fo.import_type == FileOperator.ImportType.IMPORT || 
			lm.fo.import_type == FileOperator.ImportType.PLAYLIST)
				label = "Adding files to library...";
			
			topDisplay.set_label_markup(label);
			
			return;
		}

		if(!lm.media_active) {
			topDisplay.set_label_markup("");
			debug ("setting info label as ''\n");
			return;
		}

		string beg = "";
		if(lm.media_info.media.mediatype == Media.MediaType.STATION) // radio
			beg = "<b>" + lm.media_info.media.album_artist.replace("\n", "") + "</b>\n";

		//set the title
		Media s = lm.media_info.media;
		var title = "<b>" + s.title + "</b>";
		var artist = ((s.artist != "" && s.artist != _("Unknown Artist")) ? (_(" by ") + "<b>" + s.artist + "</b>") : "");
		var album = ((s.album != "" && s.album != _("Unknown Album")) ? (_(" on ") + "<b>" + s.album + "</b>") : "");

		var media_label = beg + title + artist + album;
		topDisplay.set_label_markup(media_label);
	}

	/** This should be used whenever a call to play a new media is made
	 * @param s The media that is now playing
	 */
	public virtual void media_played(Media m, Media? old) {
		/*if(old == -2 && i != -2) { // -2 is id reserved for previews
			Media s = settings.getLastMediaPlaying();
			s = lm.media_from_name(s.title, s.artist);

			if(s.rowid != 0) {
				lm.playMedia(s.rowid);
				int position = (int)settings.getLastMediaPosition();
				topDisplay.change_value(ScrollType.NONE, position);
			}

			return;
		}*/

		updateInfoLabel();

		//reset the media position
		topDisplay.set_scale_sensitivity(true);
		topDisplay.set_scale_range(0.0, lm.media_info.media.length);

		/*if(m.mediatype == 1 || m.mediatype == 2) {
			/*message("setting position to resume_pos which is %d\n", lm.media_from_id(i).resume_pos );
			Timeout.add(250, () => {
				topDisplay.change_value(ScrollType.NONE, lm.media_from_id(i).resume_pos);
				return false;
			});*
		}
		else {
			topDisplay.change_value(ScrollType.NONE, 0);
		}*/

		//if(!mediaPosition.get_sensitive())
		//	mediaPosition.set_sensitive(true);

		//reset some booleans
		tested_for_video = false;
		queriedlastfm = false;
		media_considered_played = false;
		added_to_play_count = false;
		scrobbled_track = false;
		
		update_sensitivities();

		// if radio, we can't depend on current_position_update. do that stuff now.
		if(lm.media_info.media.mediatype == Media.MediaType.STATION) {
			queriedlastfm = true;
			
			lm.lfm.fetchCurrentAlbumInfo();
			lm.lfm.fetchCurrentArtistInfo();
			lm.lfm.fetchCurrentTrackInfo();
			lm.lfm.postNowPlaying();
			
			// always show notifications for the radio, since user likely does not know media
			mkl.showNotification(lm.media_info.media.rowid);
		}
		else {
			Timeout.add(3000, () => {
				if(lm.media_info.media != null && lm.media_info.media == m && m.rowid != LibraryManager.PREVIEW_MEDIA_ID) {
					lm.lfm.fetchCurrentSimilarSongs();
				}
				
				return false;
			});
		}
	}

	public virtual void playback_stopped(int was_playing) {
		//reset some booleans
		tested_for_video = false;
		queriedlastfm = false;
		media_considered_played = false;
		added_to_play_count = false;

		update_sensitivities();

		debug ("stopped\n");
	}

	public virtual void medias_updated(Collection<int> ids) {
		if(lm.media_active && ids.contains(lm.media_info.media.rowid)) {
			updateInfoLabel();
		}
	}

	public virtual void previousClicked () {
		if(lm.player.getPosition() < 5000000000 || (lm.media_active && lm.media_info.media.mediatype == Media.MediaType.STATION)) {
			Media prev = lm.getPrevious(true);

			/* test to stop playback/reached end */
			if(prev == null) {
				lm.player.pause();
				lm.playing = false;
				update_sensitivities();
				return;
			}
		}
		else
			topDisplay.change_value(ScrollType.NONE, 0);
	}
	
	public virtual void playClicked () {
		if(!lm.media_active) {
			debug("No media is currently playing. Starting from the top\n");
			//set current medias by current view
			Widget w = get_current_view_wrapper ();
			
			if(w is ViewWrapper) {
				((ViewWrapper)w).list_view.set_as_current_list();
			}
			else {
				w = sideTree.getWidget(sideTree.library_music_iter);
				((ViewWrapper)w).list_view.set_as_current_list();
			}

			lm.getNext(true);

			lm.playing = true;
			playButton.set_stock_id(Gtk.Stock.MEDIA_PAUSE);
			lm.player.play();
		}
		else {
			if(lm.playing) {
				lm.playing = false;
				lm.player.pause();
				
				playButton.set_stock_id(Gtk.Stock.MEDIA_PLAY);
			}
			else {
				lm.playing = true;
				lm.player.play();
				playButton.set_stock_id(Gtk.Stock.MEDIA_PAUSE);
			}
		}

		playPauseChanged();
	}

	public virtual void nextClicked() {
		// if not 90% done, skip it
		if(!added_to_play_count) {
			lm.media_info.media.skip_count++;

			// don't update, it will be updated eventually
			//lm.update_media(lm.media_info.media, false, false);
		}

		Media next;
		if(lm.next_gapless_id != 0) {
			next = lm.media_from_id(lm.next_gapless_id);
			lm.playMedia(next, false);
		}
		else
			next = lm.getNext(true);

		/* test to stop playback/reached end */
		if(next == null) {
			lm.player.pause();
			lm.playing = false;
			update_sensitivities();
			return;
		}
	}

	public virtual void loveButtonClicked() {
		lm.lfm.loveTrack(lm.media_info.media.title, lm.media_info.media.artist);
	}

	public virtual void banButtonClicked() {
		lm.lfm.banTrack(lm.media_info.media.title, lm.media_info.media.artist);
	}

	public virtual void searchFieldIconPressed(EntryIconPosition p0, Gdk.Event p1) {
		Widget w = get_current_view_wrapper ();
		w.focus(DirectionType.UP);
	}

	 void on_quit() {
		lm.settings.setLastMediaPosition((int)((double)lm.player.getPosition()/1000000000));
		if(lm.media_active) {
			lm.media_info.media.resume_pos = (int)((double)lm.player.getPosition()/1000000000);
			lm.update_media(lm.media_info.media, false, false, false);
		}
		lm.player.pause();

		// Terminate Libnotify
		Notify.uninit ();
		
		// Save UI Information
		settings.setWindowWidth(get_allocated_width());
		settings.setWindowHeight(get_allocated_height());
		settings.setSearchString (searchField.get_text());
		settings.setMoreWidth(infoPanel.get_allocated_width());
		settings.setSidebarWidth(sourcesToMedias.position);
	}


	public virtual void fileImportMusicClick() {
		if(!lm.doing_file_operations()) {
			/*if(!(GLib.File.new_for_path(lm.settings.getMusicFolder()).query_exists() && lm.settings.getCopyImportedMusic())) {
				var dialog = new MessageDialog(this, DialogFlags.DESTROY_WITH_PARENT, MessageType.ERROR, ButtonsType.OK,
				"Before importing, you must mount your music folder.");

				var result = dialog.run();
				dialog.destroy();

				return;
			}*/

			GLib.SList<File> folders = new GLib.SList<File>();
			var file_chooser = new FileChooserDialog (_("Import Music"), this,
									  FileChooserAction.SELECT_FOLDER,
									  Gtk.Stock.CANCEL, ResponseType.CANCEL,
									  Gtk.Stock.OPEN, ResponseType.ACCEPT);
			file_chooser.set_local_only(true);
			file_chooser.set_select_multiple(true); // allow user to select multiple folders

			if (file_chooser.run () == ResponseType.ACCEPT) {
				folders = file_chooser.get_files();
			}
			file_chooser.destroy ();
			
			bool tried_import_from_music_folder = false;
			var final_folders = new LinkedList<File>();
			if(folders.length() > 0) {
				foreach(var folder in folders) {
					if(!folder.get_path().has_prefix(settings.getMusicFolder())) {
						final_folders.add(folder);
					}
					else {
						tried_import_from_music_folder = true;
					}
				}
				
				if(final_folders.size > 0 && GLib.File.new_for_path(lm.settings.getMusicFolder()).query_exists()) {
					topDisplay.show_progressbar();

					lm.add_folder_to_library(final_folders);
					update_sensitivities();
				}
				else if(!GLib.File.new_for_path(lm.settings.getMusicFolder()).query_exists()) {
					doAlert("Import Failed", "Your library's music folder could not be found. Please make sure your music folder is mounted before importing.");
				}
			}
			if(tried_import_from_music_folder && final_folders.size > 0) {
				doAlert("Partial Import", "Some of the folders you selected to import are already in your Music Folder. Please rescan instead.");
			}
		}
		else {
			debug("Can't add to library.. already doing file operations\n");
		}
	}

	public virtual void fileRescanMusicFolderClick() {
		if(!lm.doing_file_operations()) {
			if(GLib.File.new_for_path(this.settings.getMusicFolder()).query_exists()) {
				topDisplay.set_label_markup("<b>" + _("Rescanning music folder for changes") + "</b>");
				topDisplay.show_progressbar();

				lm.rescan_music_folder();
				update_sensitivities();
			}
			else {
				doAlert(_("Could not find Music Folder"), _("Please make sure that your music folder is accessible and mounted."));
			}
		}
		else {
			debug("Can't rescan.. doing file operations already\n");
		}
	}

	public void resetSideTree(bool clear_views) {
		sideTree.resetView();

		// clear all other playlists, reset to Music, populate music
		if(clear_views) {
			message("clearing all views...\n");
			mainViews.get_children().foreach( (w) => {
				if(w is ViewWrapper && !(w is DeviceViewWrapper)) {
					ViewWrapper vw = (ViewWrapper)w;
					debug("doing clear\n");
					//vw.do_update(vw.current_view, new LinkedList<int>(), true, true, false);
					vw.set_media(new LinkedList<Media>());
					debug("cleared\n");
				}
			});
			message("all cleared\n");
		}
		else {
			ViewWrapper vw = (ViewWrapper)sideTree.getWidget(sideTree.library_music_iter);
			vw.set_media(lm.songs());

			vw = (ViewWrapper)sideTree.getWidget(sideTree.library_podcasts_iter);
			vw.set_media(lm.podcasts());

			vw = (ViewWrapper)sideTree.getWidget(sideTree.network_radio_iter);
			vw.set_media(lm.stations());
		}
	}

	public virtual void musicCounted(int count) {
		debug ("found %d medias, importing.\n", count);
	}

	/* this is after setting the music library */
	public virtual void musicAdded(LinkedList<Media> new_medias, LinkedList<string> not_imported) {

		if(lm.media_active) {
			updateInfoLabel();
		}
		else
			topDisplay.set_label_text("");

		//resetSideTree(false);
		//var init = searchField.get_text();
		//searchField.set_text("up");

		if(not_imported.size > 0) {
			NotImportedDialog nim = new NotImportedDialog(this, not_imported, lm.settings.getMusicFolder());
			nim.show();
		}

		update_sensitivities();

		//now notify user
		try {
			if (Notify.is_initted () && new_medias.size > 0) {
				notification.close();
				
				notification.update(_("Import Complete"), _("BeatBox has imported your library."), "beatbox");
				notification.set_image_from_pixbuf(Icons.BEATBOX.render (Gtk.IconSize.DIALOG));
				
				notification.show();
				notification.set_timeout (Notify.EXPIRES_DEFAULT);
			}
		}
		catch(GLib.Error err) {
			stderr.printf("Could not show notification: %s\n", err.message);
		}
	}

	/* this is when you import music from a foreign location into the library */
	public virtual void musicImported(LinkedList<Media> new_medias, LinkedList<string> not_imported) {
		if(lm.media_active) {
			updateInfoLabel();
		}
		else
			topDisplay.set_label_text("");

		resetSideTree(false);
		//searchField.changed();

		update_sensitivities();
	}

	public virtual void musicRescanned(LinkedList<Media> new_medias, LinkedList<string> not_imported) {
		if(lm.media_active) {
			updateInfoLabel();
		}
		else
			topDisplay.set_label_text("");

		resetSideTree(false);
		//searchField.changed();
		debug("music Rescanned\n");
		update_sensitivities();
	}

	public virtual void medias_removed(LinkedList<int> removed) {
		update_sensitivities();
	}

	public void editEqualizerClick() {
		EqualizerWindow ew = new EqualizerWindow(lm, this);
		ew.show();
	}

	public void editPreferencesClick() {
		PreferencesWindow pw = new PreferencesWindow(lm, this);

		pw.changed.connect( (folder) => {
			setMusicFolder(folder);
		});
	}

	public void setMusicFolder(string folder) {
		if(lm.doing_file_operations())
			return;

		if(lm.song_ids().size > 0 || lm.playlist_count() > 0) {
			var smfc = new SetMusicFolderConfirmation(lm, this, folder);
			smfc.finished.connect( (cont) => {
				if(cont) {
					lm.set_music_folder(folder);
				}
			});
		}
		else {
			lm.set_music_folder(folder);
		}
	}

	public virtual void end_of_stream() {
		nextClicked();
	}

	public virtual void current_position_update(int64 position) {
		if (!lm.media_active || !initializationFinished)
			return;

		if (lm.media_info.media.rowid == Media.PREVIEW_ROWID) // is preview
			return;

		double sec = ((double)position/1000000000);

		if(lm.player.set_resume_pos)
			lm.media_info.media.resume_pos = (int)sec;

		// at about 3 seconds, update last fm. we wait to avoid excessive querying last.fm for info
		if(position > 3000000000 && !queriedlastfm) {
			queriedlastfm = true;
			
			lm.lfm.fetchCurrentSimilarSongs();
			lm.lfm.fetchCurrentAlbumInfo();
			lm.lfm.fetchCurrentArtistInfo();
			lm.lfm.fetchCurrentTrackInfo();
			lm.lfm.postNowPlaying();
		}

		//at 30 seconds in, we consider the media as played
		if(position > 30000000000 && !media_considered_played) {
			media_considered_played = true;
			lm.media_info.media.last_played = (int)time_t();

			if(lm.media_info.media.mediatype == Media.MediaType.PODCAST) { //podcast
				added_to_play_count = true;
				++lm.media_info.media.play_count;
			}

			lm.update_media(lm.media_info.media, false, false, true);

			// add to the already played list
			lm.add_already_played(lm.media_info.media);

#if HAVE_ZEITGEIST
			var event = new Zeitgeist.Event.full (Zeitgeist.ZG_ACCESS_EVENT,
			                                       Zeitgeist.ZG_SCHEDULED_ACTIVITY, "app://beatbox.desktop",
			                                       new Zeitgeist.Subject.full(lm.media_info.media.uri,
			                                                                   Zeitgeist.NFO_AUDIO,
			                                                                   Zeitgeist.NFO_FILE_DATA_OBJECT,
			                                                                   "text/plain", "",
			                                                                   lm.media_info.media.title, ""));
			new Zeitgeist.Log ().insert_events_no_reply(event);
#endif
		}

		// at halfway, scrobble
		if((double)(sec/(double)lm.media_info.media.length) > 0.50 && !scrobbled_track) {
			scrobbled_track = true;
			
			lm.lfm.postScrobbleTrack();
		}

		// at 80% done with media, add 1 to play count
		if((double)(sec/(double)lm.media_info.media.length) > 0.80 && !added_to_play_count) {
			added_to_play_count = true;
			lm.media_info.media.play_count++;
			lm.update_media(lm.media_info.media, false, false, true);
		}
	}

	public void media_not_found(int id) {
		file_not_found_pileup.add(lm.media_from_id(id));
		infoBarLabel.set_markup("<b>" + file_not_found_pileup.size.to_string() + "</b> media files could not be found.");
		infoBar.show_all();
	}
	
	public void media_found(int id) {
		file_not_found_pileup.remove(lm.media_from_id(id));
		infoBarLabel.set_markup("<b>" + file_not_found_pileup.size.to_string() + "</b> media files could not be found.");
		infoBar.set_visible(file_not_found_pileup.size > 0);
	}

	public void set_statusbar_info (ViewWrapper.Hint media_type, uint total_medias,
									 uint total_mbs, uint total_seconds) {
		statusBar.set_total_medias (total_medias, media_type);
		statusBar.set_files_size (total_mbs);
		statusBar.set_total_time (total_seconds);
	}

	void repeatChooserOptionChanged(int val) {
		if(val == 0)
			lm.set_repeat_mode(LibraryManager.Repeat.OFF);
		else if(val == 1)
			lm.set_repeat_mode(LibraryManager.Repeat.MEDIA);
		else if(val == 2)
			lm.set_repeat_mode(LibraryManager.Repeat.ALBUM);
		else if(val == 3)
			lm.set_repeat_mode(LibraryManager.Repeat.ARTIST);
		else if(val == 4)
			lm.set_repeat_mode(LibraryManager.Repeat.ALL);
	}

	void shuffleChooserOptionChanged(int val) {
		if(val == 0)
			lm.set_shuffle_mode(LibraryManager.Shuffle.OFF);
		else if(val == 1)
			lm.set_shuffle_mode(LibraryManager.Shuffle.ALL);
		
	}

	bool addPlaylistChooserOptionClicked(Gdk.EventButton event) {
		if (event.button == 1) {
			sideTree.playlistMenuNewClicked();
			return true;
		}
		else if(event.button == 3) {
			newPlaylistMenu.show_all();
			newPlaylistMenu.popup (null, null, null, 3, get_current_event_time());
		}

		return true;
	}

	public virtual void infoPanelChooserOptionChanged(int val) {
		infoPanel.set_visible(val == 1);
		lm.settings.setMoreVisible(val == 1);
	}

	public Widget? get_current_view_wrapper () {
		return mainViews.get_nth_page (mainViews.get_current_page());
	}

	public void searchFieldActivate() {
		Widget w = get_current_view_wrapper ();

		if(w is ViewWrapper) {
			ViewWrapper vw = (ViewWrapper)w;
			
			if (vw.have_list_view)
				vw.list_view.set_as_current_list ();
			
			lm.current_index = 0;
			lm.playMedia(lm.media_from_id(lm.mediaFromCurrentIndex(0)), false);

			if(!lm.playing)
				playClicked();
		}
	}

	public virtual void dragReceived(Gdk.DragContext context, int x, int y, Gtk.SelectionData data, uint info, uint timestamp) {
		if(dragging_from_music)
			return;

		var files_dragged = new LinkedList<string>();
		debug("dragged\n");
		foreach (string uri in data.get_uris ()) {
			files_dragged.add(File.new_for_uri(uri).get_path());
		}

		lm.add_files_to_library(files_dragged);
	}

	public void doAlert(string title, string message) {
		var dialog = new MessageDialog(this, DialogFlags.MODAL, MessageType.ERROR, ButtonsType.OK,
				title);

		dialog.title = "BeatBox";
		dialog.secondary_text = message;
		dialog.secondary_use_markup = true;

		dialog.run();
		dialog.destroy();
	}
	
	void info_bar_resonse(int response_id) {
		if(response_id == Gtk.ResponseType.YES) {
			FileNotFoundDialog fnfd = new FileNotFoundDialog(lm, this, file_not_found_pileup);
			fnfd.present();
			
			file_not_found_pileup = new LinkedList<Media>();
		}
		
		infoBar.hide();
	}
	
	public void show_video_mode() {
		message("a\n");
		if(viewSelector.get_children().length() != 3) {
			stdout.printf("turning on video\n");
			viewSelector.append(Icons.VIEW_VIDEO.render_image (Gtk.IconSize.MENU));
			viewSelector.selected = 3;
		}
	}
	
	public void hide_video_mode() {
		message("b\n");
		stdout.printf("turning off video\n");
		if(viewSelector.selected == 2) {
			viewSelector.selected = 1; // show list
		}
		
		if(viewSelector.get_children().length() == 3) {
			viewSelector.remove(2);
		}
	}
}

