/*-
 * Copyright (c) 2011-2012       Scott Ringwelski <sgringwe@mtu.edu>
 *
 * Originally Written by Scott Ringwelski for BeatBox Music Player
 * BeatBox Music Player: http://www.launchpad.net/beat-box
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authored by: Scott Ringwelski <sgringwe@mtu.edu>
 *              Victor Eduardo <victoreduardm@gmail.com>
 */

using Gtk;
using Granite.Widgets;
using Gee;

public class BeatBox.ViewWrapper : VBox {
	public LibraryManager lm { get; private set; }
	public LibraryWindow  lw { get; private set; }

	/* MAIN WIDGETS (VIEWS) */
	HBox				 top_area		{ get; private set; }
	public Granite.Widgets.ModeButton	 selector		{ get; private set; }
	ScrolledWindow 		 list_scroll    { get; private set; }
	ScrolledWindow		 album_scroll	{ get; private set; }
//	ScrolledWindow		 artist_scroll	{ get; private set; }
//	ScrolledWindow		 genre_scroll	{ get; private set; }
	public GenericList	 list_view      { get; private set; }
	public GenericGrid	 album_view		{ get; private set; }
//	public GenericGrid	 artist_view	{ get; private set; }
//	public GenericGrid	 genre_view		{ get; private set; }
	public WarningLabel  error_box      { get; private set; }
	public Welcome       welcome_screen { get; private set; }
	public PopupListView popup			{ get; private set; }

	private Notebook view_container; // Wraps all the internal views for super fast switching

	public enum ViewType {
		LIST    = 1,
		ALBUM	= 2,
		ARTIST	= 3,
		GENRE 	= 4,
		ERROR   = 5, // For error boxes
		WELCOME = 6, // For welcome screens
		NONE    = 7  // Custom views
	}
	
	/**
	 * This is by far the most important property of this object.
	 * It defines how child widgets behave and some other properties.
	 */
	public enum Hint {
		NONE,
		MUSIC,
		PODCAST,
		AUDIOBOOK,
		STATION,
		SIMILAR,
		QUEUE,
		HISTORY,
		PLAYLIST,
		SMART_PLAYLIST,
		CDROM,
		DEVICE_AUDIO,
		DEVICE_PODCAST,
		DEVICE_AUDIOBOOK,
		ALBUM_LIST;
	}

	public Hint hint { get; private set; }
	public int relative_id { get; private set; }
	public string media_representation { get; private set; }

	public int index { get { return lw.mainViews.page_num(this); } }
	public bool is_current_wrapper {
		get {
			return (lw.initializationFinished ? (index == lw.mainViews.get_current_page()) : false);
		}
	}


	/* UI PROPERTIES */
	public bool have_list_view       { get { return list_view != null;      } }
	public bool have_album_view		 { get { return album_view != null;		} }
//	public bool have_artist_view	 { get { return artist_view != null;	} }
//	public bool have_genre_view		 { get { return genre_view != null;	} }
	public bool have_error_box       { get { return error_box != null;      } }
	public bool have_welcome_screen  { get { return welcome_screen != null; } }
	
	/**
	 * This boolean is extremely important. It defines whether we show the views or not,
	 * and also other widgets, like the error box or welcome screen.
	 */
	bool needs_update;
	public bool have_media { get { return media_count > 0; } }
	Collection<Media> original_medias;
	HashTable<int, int> current_medias; // rowid, int
	
	bool album_needs_update;
	Gee.HashMap<string, Album> album_keys; // helper
	
//	bool artist_needs_update;
//	Gee.HashMap<string, Artist> artist_keys; // helper
	
//	bool genre_needs_update;
//	Gee.HashMap<string, Genre> genre_keys; // helper

	/**
	 * MEDIA DATA
	 *
	 * These data structures hold information about the media shown in the views.
	 **/

	// ALL the media. Data source.
	protected Mutex in_update;
	public int media_count { get { return (list_view != null) ? (int)list_view.get_table().size() : 0; } }
	protected string last_search = "";
	protected string _artist_filter = "";
	protected string _album_filter = "";
	protected string _genre_filter = "";

	// Stops from searching unnecesarilly when changing b/w 0 words and search.
	private bool showing_all { get { return list_view.get_visible_table().size() == list_view.get_table().size(); } }
	public int showing_media_count { get { return (list_view != null) ? (int)list_view.get_visible_table().size() : 0; } }
	bool initialized;
	
	public string album_filter {
		get {
			return _album_filter;
		}
		set {
			if(_album_filter != value) {
				_album_filter = value;
				
				re_search();
			}
		}
	}
	
	public string artist_filter {
		get {
			return _artist_filter;
		}
		set {
			if(_artist_filter != value) {
				_artist_filter = value;
				
				re_search();
			}
		}
	}
	
	public string genre_filter {
		get {
			return _genre_filter;
		}
		set {
			if(_genre_filter != value) {
				_genre_filter = value;
				
				re_search();
			}
		}
	}
	
	bool multiple_grid_views {
		get {
			/*if((have_album_view && have_artist_view) || 
			(have_album_view && have_genre_view) ||
			(have_artist_view && have_genre_view)) {
				return true;
			}*/
			
			return false;
		}
	}
	
	public string grid_artist_filter { get; set; }
	public string grid_genre_filter { get; set; }
	
	public ViewWrapper (LibraryWindow lw, Collection<Media> the_medias, TreeViewSetup tvs, int id)
	{
		initialized = false;

		this.lm = lw.lm;
		this.lw = lw;
		this.hint = tvs.get_hint();
		this.relative_id = id;
		original_medias = the_medias;
		
		current_medias = new HashTable<int, int>(null, null); // rowid, int
		album_keys = new HashMap<string, Album>(null, null);
//		artist_keys = new HashMap<string, Artist>(null, null);
//		genre_keys = new HashMap<string, Genre>(null, null);
		
		_album_filter = "";
		_artist_filter = "";
		_genre_filter = "";
		
		in_update = new Mutex();
		
		top_area = new HBox(true, 0);
		selector = new Granite.Widgets.ModeButton();
		var trans_box1 = new Gtk.EventBox();
		trans_box1.set_visible_window(false);
		var trans_box2 = new Gtk.EventBox();
		trans_box1.set_visible_window(false);
		top_area.pack_start(trans_box1, false, true, 0);
		top_area.pack_start(selector, false, true, 0);
		top_area.pack_start(trans_box2, false, true, 0);
		top_area.get_style_context().add_class(STYLE_CLASS_PRIMARY_TOOLBAR);
		
		// Setup view container
		view_container = new Notebook ();
		view_container.show_tabs = false;
		view_container.show_border = false;
		
		//this.pack_start (top_area, false, true, 0);
		this.pack_end (view_container, true, true, 0);

		switch (hint) {
			case Hint.MUSIC:
				media_representation = "song";
				list_view = new MusicList (lm, tvs, id);
				album_view = new AlbumGrid(lm, tvs, id);
				//artist_view = new ArtistGrid(lm, tvs, id);
				//genre_view = new GenreGrid(lm, tvs, id);
				error_box = new WarningLabel();
				welcome_screen = new Granite.Widgets.Welcome(_("Get Some Tunes"), _("BeatBox can't find your music."));
				
				break;
			case Hint.SIMILAR:
				media_representation = "song";
				
				list_view = new MusicList(lm, tvs, id);

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.QUEUE:
				media_representation = "item";
				
				list_view = new MusicList(lm, tvs, id);

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.HISTORY:
				media_representation = "item";
				
				list_view = new MusicList(lm, tvs, id);

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.PODCAST:
				media_representation = "podcast";
				
				list_view = new PodcastList (lm, tvs, id);
				album_view = new AlbumGrid(lm, tvs, id);
				//artist_view = new ArtistGrid(lm, tvs, id);
				//genre_view = new GenreGrid(lm, tvs, id);
				welcome_screen = new Granite.Widgets.Welcome(_("Subscribe to Podcasts"), _("No Podcasts were found."));

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.DEVICE_PODCAST:
				media_representation = "podcast";
				
				list_view = new PodcastList (lm, tvs, id);

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.STATION:
				media_representation = "station";
				
				list_view = new RadioList(lm, tvs, id);
				welcome_screen = new Granite.Widgets.Welcome(_("Turn up the Radio"), _("No Stations were found."));

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.AUDIOBOOK:
				list_view = new MusicList(lm, tvs, id);

				break;
			case Hint.DEVICE_AUDIOBOOK:
				list_view = new MusicList(lm, tvs, id);

				break;
			case Hint.CDROM:
				media_representation = "song";
				
				list_view = new MusicList(lm, tvs, id);

				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.PLAYLIST:
				media_representation = "item";
				
				list_view = new MusicList(lm, tvs, id);
				album_view = new AlbumGrid(lm, tvs, id);
				//artist_view = new ArtistGrid(lm, tvs, id);
				//genre_view = new GenreGrid(lm, tvs, id);
				
				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			case Hint.SMART_PLAYLIST:
				media_representation = "item";
				
				list_view = new MusicList(lm, tvs, id);
				album_view = new AlbumGrid(lm, tvs, id);
				//artist_view = new ArtistGrid(lm, tvs, id);
				//genre_view = new GenreGrid(lm, tvs, id);
				
				error_box = new WarningLabel();
				set_default_warning ();
				
				break;
			default:
				media_representation = "item";
				
				list_view = new MusicList(lm, tvs, id);

				break;
		}

		// Now setup the view wrapper based on available widgets

		if (have_error_box) {
			view_container.append_page (error_box);
		}

		if (have_welcome_screen) {
			view_container.append_page (welcome_screen);
		}

		if (have_list_view) {
			list_view.set_parent_wrapper (this);
			
			list_scroll = new ScrolledWindow(null, null);
			list_scroll.set_policy(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
			list_scroll.add(list_view);
			view_container.append_page (list_scroll);
			
			// Set sort data from saved session
			list_view.set_sort_column_id(tvs.sort_column_id, tvs.sort_direction);
		}
		
		if(have_album_view) {
			album_view.set_parent_wrapper (this);
			
			album_scroll = new ScrolledWindow(null, null);
			album_scroll.set_policy(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
			album_scroll.add(album_view);
			view_container.append_page (album_scroll);
			
			selector.append(new Label("Albums"));
			selector.selected = 0;
			
			popup = new PopupListView(lm, this);
		}
		
/*		if(have_artist_view) {
			artist_view.set_parent_wrapper (this);
			
			artist_scroll = new ScrolledWindow(null, null);
			artist_scroll.set_policy(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
			artist_scroll.add(artist_view);
			view_container.append_page (artist_scroll);
			
			selector.append(new Label("Artists"));
		}
		
		if(have_genre_view) {
			genre_view.set_parent_wrapper (this);
			
			genre_scroll = new ScrolledWindow(null, null);
			genre_scroll.set_policy(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
			genre_scroll.add(genre_view);
			view_container.append_page (genre_scroll);
			
			selector.append(new Label("Genres"));
		}
*/
		
		// Populate views
		do_initial_populate();

		lm.medias_updated.connect (medias_updated);
		lm.medias_added.connect (medias_added);
		lm.medias_removed.connect (medias_removed);
		
		// Listen for playlist additions/removals
		if(hint == ViewWrapper.Hint.PLAYLIST) {
			lm.playlist_from_id(relative_id).changed.connect(playlist_changed);
		}
		else if(hint == ViewWrapper.Hint.QUEUE) {
			lm.queue_changed.connect(queue_changed);
		}
		else if(hint == ViewWrapper.Hint.HISTORY) {
			lm.already_played_changed.connect(already_played_changed);
		}
		
		lw.viewSelector.mode_changed.connect(main_view_selector_changed);
		selector.mode_changed.connect(view_selector_changed);
		lw.searchField.changed.connect (search_field_changed);
		lw.shuffleChooser.option_changed.connect(shuffleChooserOptionChanged);

		initialized = true;
	}

	public ViewWrapper.with_view (Gtk.Widget view) {
		view_container.append_page (view);
		this.hint = Hint.NONE;
		set_active_view (ViewType.NONE);
		update_library_window_widgets ();
	}

	// We only check for white space at the moment
	private bool get_is_valid_search_string (string s) {
		if (s.length < 1)
			return true;

		int white_space = 0;
		unichar c;

		for (int i = 0; s.get_next_char (ref i, out c);)
			if (c.isspace())
				++ white_space;

		if (white_space == s.length) {
			debug ("STRING '%s' IS WHITESPACE", s);
			return false;
		}

		return true;
	}
	
	void main_view_selector_changed() {
		if(lw.viewSelector.selected == 0) {
			set_active_view(ViewType.LIST);
			top_area.hide();
		}
		else if(lw.viewSelector.selected == 1) {
			top_area.set_visible(multiple_grid_views);
			view_selector_changed();
		}
	}
	
	void view_selector_changed() {
		if(selector.selected == 0)
			set_active_view(ViewType.ALBUM);
		else if(selector.selected == 1)
			set_active_view(ViewType.ARTIST);
		else if(selector.selected == 2)
			set_active_view(ViewType.GENRE);
	}
	
	/**
	 * Convenient visibility method
	 */
	protected void set_active_view (ViewType type, out bool successful = null) {
		int view_index = -1;
		
		if(type == ViewType.LIST || type == ViewType.ALBUM || type == ViewType.ARTIST) {
			if(needs_update)
				set_media(original_medias, true);
		}

		// Find position in notebook
		switch (type) {
			case ViewType.LIST:
				if (have_list_view)
					view_index = view_container.page_num (list_scroll);
				break;
			case ViewType.ALBUM:
				if (have_album_view) {
					//if(album_needs_update)
					//	set_albums();
					
					view_index = view_container.page_num (album_scroll);
				}
				break;
			case ViewType.ARTIST:
				/*if(have_artist_view) {
					view_index = view_container.page_num (artist_scroll);
				}*/
				break;
			case ViewType.GENRE:
				/*if(have_genre_view) {
					view_index = view_container.page_num (genre_scroll);
				}*/
				break;
			case ViewType.ERROR:
				if (have_error_box)
					view_index = view_container.page_num (error_box);
				break;
			case ViewType.WELCOME:
				if (have_welcome_screen)
					view_index = view_container.page_num (welcome_screen);
				break;
		}

		// i.e. we're not switching the view if it is not available
		if (view_index < 0) {
			successful = false;
			return;
		}

		// Set view as current
		view_container.set_current_page (view_index);
		
		// If showing grid, give all grid options if more than album
		if(type == ViewType.ALBUM || type == ViewType.ARTIST || type == ViewType.GENRE)
			top_area.set_visible(multiple_grid_views);
		else
			top_area.set_visible(false);

		// Update BeatBox's toolbar widgets
		if(visible)
			update_library_window_widgets ();

		successful = true;
	}

	/**
	 * This method ensures that the view switcher and search box are sensitive/insensitive when they have to.
	 * It also selects the proper view switcher item based on the current view.
	 */
	protected void update_library_window_widgets () {
		if (!is_current_wrapper)
			return;
		
		lw.viewSelector.set_sensitive((have_media && have_album_view) || selector.get_children().length() == 3);
		lw.searchField.set_sensitive (have_media);
		
		lw.update_sensitivities ();
	}

	public void play_first_media () {
		if (!have_list_view)
			return;
		
		lm.media_info.media = null;
		list_view.set_as_current_list();
		lm.playMedia (lm.media_from_id(lm.mediaFromCurrentIndex(0)), false);

		if(!lm.playing)
			lw.playClicked();
	}

	/**
	 * This handles updating all the shared stuff outside the view area.
	 *
	 * You should only call this method on the respective ViewWrapper whenever the sidebar's
	 * selected view changes.
	 *
	 * Note: The sidebar-item selection and other stuff is handled automatically by the LibraryWindow
	 *       by request of SideTreeView. See LibraryManager :: set_active_view() for more details.
	 */
	public void set_as_current_view () {
		// Restore this view wrapper's search string
		lw.search_suggester.set_listen_for_search(false);
		lw.searchField.set_text (last_search);
		lw.search_suggester.set_listen_for_search(true);
		
		update_library_window_widgets ();
		set_statusbar_info ();
		check_have_media ();
		
		list_view.scroll_to_current_media (false);
	}
	
	void set_default_warning () {
		if(!have_error_box)
			return;
		
		switch (hint) {
			case Hint.MUSIC:
				
				break;
			case Hint.SIMILAR:
				// Leave this to SimilarViewWrapper
				
				break;
			case Hint.QUEUE:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No songs in Queue") + "</span>\n\n" + _("To queue a song, drag and drop a song from a list onto the queue sidebar item.") + "\n" + _("When a song finishes, the queued songs will be played first before the next song in the currently playing list."));
				
				break;
			case Hint.HISTORY:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No songs in History") + "</span>\n\n" + _("After a part of a song has been played, it is added to the history list.") + "\n" + _("You can use this list to see all the songs you have played during the current session."));
				
				break;
			case Hint.PODCAST:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No Podcasts Found") + "</span>\n\n" + _("To add a podcast, visit a website such as Miro Guide to find RSS Feeds.") + "\n" + _("You can then copy and paste the feed into the \"Add Podcast\" window by right clicking on \"Podcasts\"."));

				break;
			case Hint.DEVICE_PODCAST:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No Podcasts Found") + "</span>\n\n" + _("To add a podcast, visit a website such as Miro Guide to find RSS Feeds.") + "\n" + _("You can then copy and paste the feed into the \"Add Podcast\" window by right clicking on \"Podcasts\"."));

				break;
			case Hint.STATION:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No Internet Radio Stations Found") + "</span>\n\n" + _("To add a station, visit a website such as SomaFM to find PLS or M3U files.") + "\n" + _("You can then import the file to add the station."));
				
				break;
			case Hint.AUDIOBOOK:
			
				break;
			case Hint.DEVICE_AUDIOBOOK:
			
				break;
			case Hint.CDROM:
				error_box.show_icon = true;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("Audio CD Invalid") + "</span>\n\n" + _("BeatBox could not read the contents of this Audio CD."));
				
				break;
			case Hint.PLAYLIST:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No Media") + "</span>\n\n" + _("To add to this playlist, drag and drop songs from a list onto the sidebar item or") + "\n" + _("right click on an item and choose \"Add to Playlist\"."));
				
				break;
			case Hint.SMART_PLAYLIST:
				error_box.show_icon = false;
				error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No Media") + "</span>\n\n" + _("To add to this smart playlist, right click on it in the sidebar and edit its rules."));
				
				break;
			default:

				break;
		}
	}
	
	void set_search_warning () {
		if(!have_error_box)
			return;
		
		error_box.show_icon = false;
		error_box.setWarning ("<span weight=\"bold\" size=\"larger\">" + _("No Results") + "</span>\n\n" + _("Your search query returned no results."));
	}
	
	void check_have_media () {
		if(have_media) {
			if(list_view.get_visible_table().size() > 0) {
				if(lw.viewSelector.selected == 0 || !have_album_view) {
					set_active_view(ViewType.LIST);
				}
				else {
					if(selector.selected == 0)
						set_active_view(ViewType.ALBUM);
					else if(selector.selected == 1)
						set_active_view(ViewType.ARTIST);
					else if(selector.selected == 2)
						set_active_view(ViewType.GENRE);
				}
			}
			else {
				set_search_warning ();
				set_active_view(ViewType.ERROR);
			}
		}
		else {
			if(hint == Hint.MUSIC || hint == Hint.PODCAST || hint == Hint.STATION) {
				set_active_view(ViewType.WELCOME);
			}
			else {
				set_default_warning ();
				set_active_view(ViewType.ERROR);
			}
		}
	}

	public void set_statusbar_info() {
		if (!is_current_wrapper)
			return;

		if(showing_media_count < 1) {
			lw.set_statusbar_info (hint, 0, 0, 0);
			return;
		}

		uint count = 0;
		uint total_time = 0;
		uint total_mbs = 0;
		foreach(Media media in list_view.get_visible_table().get_values()) {
			count ++;
			total_time += media.length;
			total_mbs += media.file_size;
		}

		lw.set_statusbar_info(hint, count, total_mbs, total_time);
	}
	
	void shuffleChooserOptionChanged(int val) {
		list_view.toggle_shuffle (val == 1);
	}

	void search_field_changed() {
		if (!is_current_wrapper)
			return;

		var new_search = lw.searchField.get_text().down();

		if (!get_is_valid_search_string (new_search))
			return;
		
		if(lw.initializationFinished) {
			// Clear previous filters
			if(lw.search_suggester.get_listen_for_search())
				clear_filters ();
			
			if(have_list_view)
				list_view.do_search(new_search);
			if(have_album_view)
				album_view.do_search(new_search);
/*			if(have_artist_view)
				artist_view.do_search(new_search);
			if(have_genre_view)
				genre_view.do_search(new_search);
*/

			last_search = new_search;
			set_statusbar_info ();
			check_have_media ();
		}
	}
	
	public void re_search() {
		stdout.printf("researching %s,%s,%s\n", _album_filter, _artist_filter, _genre_filter);
		if(have_list_view)
			list_view.do_search(null);
		if(have_album_view)
			album_view.do_search(null);
/*		if(have_artist_view)
			artist_view.do_search(null);
		if(have_genre_view)
			genre_view.do_search(null);
*/
			
		update_library_window_widgets();
	}
	
	public void clear_filters () {
		_album_filter = "";
		_artist_filter = "";
		_genre_filter = "";
		
		re_search();
	}


	/**
	============================================================================
	                                  DATA STUFF
	============================================================================
	*/
	
	// For playlists
	void playlist_changed () {
		Playlist p = lm.playlist_from_id(relative_id);
		
		set_media (p.medias ());
	}
	
	// For queue
	void queue_changed () {
		set_media(lm.queue());
	}
	
	// For queue
	void already_played_changed () {
		set_media(lm.already_played());
	}
	
	/*****************************************
	 * The following three functions (media_added/updated/removed)
	 * only apply to view wrappers with songs in the library
	 * that can be referenced by an int id. Other lists, such
	 * as devices and cdroms do not use this
	 * *****************************************/
	void medias_added(LinkedList<int> ids) {
		if(!have_list_view && !have_album_view)
			return;
		
		in_update.lock (); // mutual exclusion
		
		var to_add = new LinkedList<int>();
		if(hint == Hint.MUSIC || hint == Hint.PODCAST || hint == Hint.STATION) { // See if it matches our type
			lm.do_search (ids, out to_add, null, null, null, null, hint, "");
		}
		else if(hint == Hint.SMART_PLAYLIST) { // See if it matches rules
			var to_add_medias = lm.smart_playlist_from_id(relative_id).analyze(lm, ids);
			foreach(var m in to_add_medias)
				to_add.add(m.rowid);
		}
		
		// Add all that should be added
		var to_add_list = new GLib.List<Media>();
		var to_add_grid = new GLib.List<Album>();
		Album alb;
		foreach(int i in to_add) {
			if(current_medias.get(i) == 0) {
				current_medias.set(i, 1);
				Media m = lm.media_from_id(i);
				string key = lm.album_key(m);
				to_add_list.append(lm.media_from_id(i));
				
				if((alb = album_keys.get(key)) == null) {
					alb = new Album(m.album_artist, m.album);
					album_keys.set(key, alb);
					to_add_grid.append(alb);
				}
				alb.add_media(m);
			}
		}
		
		if(have_list_view)
			list_view.add_medias(to_add_list);
		if(have_album_view)
			album_view.add_objects(to_add_grid);
		
		in_update.unlock ();
		
		update_library_window_widgets ();
		set_statusbar_info ();
		check_have_media ();
	}
	
	/**
	 * Do search to find which ones should be added, removed from this particular view
	 * does not re-anaylyze smart playlist_views or playlist_views.
	 */
	public void medias_updated(LinkedList<int> ids, bool metadata_changed) {
		if(!have_list_view && !have_album_view)
			return;
		
		// These views should not react to updates, only queues/plays/add to playlist/remove
		if(hint == Hint.SIMILAR || hint == Hint.QUEUE ||
		hint == Hint.HISTORY || hint == Hint.PLAYLIST)
			return;
			
		// We don't want to mix with device lists
		if(hint == Hint.CDROM || hint == Hint.DEVICE_AUDIO ||
		hint == Hint.DEVICE_AUDIOBOOK || hint == Hint.DEVICE_PODCAST)
			return;
		
		// for large updates (lots of media), go with "refresh" style.
		// this is faster, but refreshes the table losing selections
		if(ids.size > 500) {
			medias_removed (ids);
			medias_added (ids);
		}
		
		// otherwise, do slower but smoother update
		else {
			bool update_album_view = (have_album_view && album_view.visible);
			in_update.lock ();
			
			var should_be = new LinkedList<int>();
			if(hint == Hint.SMART_PLAYLIST) {
				var should_be_medias = lm.smart_playlist_from_id(relative_id).analyze(lm, ids);
				foreach(var m in should_be_medias)
					should_be.add(m.rowid);
			}
			else if(hint == Hint.MUSIC || hint == Hint.PODCAST || hint == Hint.STATION)
				lm.do_search (ids, out should_be, null, null, null, null, hint);
			else {
				should_be = ids;
			}

			var to_add = new HashTable<Media, int>(null, null);
			var to_remove = new HashTable<Media, int>(null, null);
			
			// add elements that should be here
			foreach(int i in should_be) {
				if(current_medias.get(i) == 0) {
					to_add.set(lm.media_from_id(i), 1);
				}
			}

			// remove elements
			foreach(int i in ids) {
				if(!should_be.contains(i)) {
					to_remove.set(lm.media_from_id(i), 1);
				}
			}
			
			// if some additions or removals, combine existing, added, and new to make the new table
			if(to_remove.size() > 0 || to_add.size() > 0 || update_album_view) {
				var new_list = new HashTable<int, Media>(null, null);
				var new_grid = new HashTable<int, Album>(null, null);
				
				// reset album_keys
				album_keys = new HashMap<string, Album>(null, null);
				Album alb;
				foreach(var m in list_view.get_table().get_values()) {
					if(to_remove.get(m) == 0) {
						current_medias.set(m.rowid, 1);
						string key = lm.album_key(m);
						new_list.set((int)new_list.size(), m);
						
						if(update_album_view) {
							if((alb = album_keys.get(key)) == null) {
								alb = new Album(m.album_artist, m.album);
								album_keys.set(key, alb);
								new_grid.set((int)new_grid.size(), alb);
							}
							alb.add_media(m);
						}
					}
					else {
						current_medias.set(m.rowid, 0);
					}
				}
				foreach(var m in to_add.get_keys()) {
					if(current_medias.get(m.rowid) == 0) {
						current_medias.set(m.rowid, 1);
						string key = lm.album_key(m);
						new_list.set((int)new_list.size(), m);
						
						if(update_album_view) {
							if((alb = album_keys.get(key)) == null) {
								alb = new Album(m.album_artist, m.album);
								album_keys.set(key, alb);
								new_grid.set((int)new_grid.size(), alb);
							}
							alb.add_media(m);
						}
					}
				}
				
				if(to_remove.size() > 0 || to_add.size() > 0) {
					list_view.set_table(new_list, true);
				}
				else
					list_view.resort();
				
				if(update_album_view) {
					album_view.set_table(new_grid, true);
				}
				else
					album_needs_update = true;
			}
			else { // if same list, resort it is all.
				if(have_list_view)
					list_view.resort ();
				
				album_needs_update = true;
			}

			in_update.unlock ();
		}
		
		update_library_window_widgets ();
		set_statusbar_info ();
		check_have_media ();
	}
	
	void medias_removed(LinkedList<int> ids) {
		if(!have_list_view && !have_album_view)
			return;
		
		bool update_album_view = (have_album_view && !album_needs_update);
		
		in_update.lock ();
		var list_remove = new GLib.HashTable<Media, int>(null, null);
		var grid_remove = new Gee.HashMap<string, int>(null, null);
		Album alb;
		foreach(int i in ids) {
			current_medias.set(i, 0);
			Media m = lm.media_from_id(i);
			string key = lm.album_key(m);
			
			list_remove.set(m, 1);
			
			if(update_album_view) {
				if((alb = album_keys.get(key)) != null) {
					grid_remove.set(key, 1);
					album_keys.unset(key);
				}
			}
		}

		// Now update the views to reflect the changes
		if(have_list_view)
			list_view.remove_medias(list_remove);
		
		if(update_album_view)
			album_view.remove_objects(grid_remove);
		else
			album_needs_update = true;
		
		in_update.unlock();
		
		update_library_window_widgets ();
		set_statusbar_info ();
		check_have_media ();
	}

	public void set_media (Collection<Media> new_media, bool update_grid = false) {
		if(have_list_view || have_album_view) {
			bool update_album_view = (have_album_view && album_view.visible) || update_grid || needs_update;
			//bool update_artist_view = (have_artist_view && artist_view.visible) || update_grid || needs_update;
			//bool update_genre_view = (have_genre_view && genre_view.visible) || update_grid || needs_update;
			
			in_update.lock ();
			
			current_medias = new HashTable<int, int>(null, null); // rowid, int
			var media_list = new HashTable<int, Media>(null, null);
			var album_list = new HashTable<int, Album>(null, null);
			//var artist_list = new HashTable<int, Artist>(null, null);
			//var genre_list = new HashTable<int, Genre>(null, null);
			
			if(update_album_view) {
				album_keys = new Gee.HashMap<string, Album>();
			}
			/*if(update_artist_view) {
				artist_keys = new Gee.HashMap<string, Artist>();
			}
			if(update_genre_view) {
				genre_keys = new Gee.HashMap<string, Genre>();
			}*/
			
			Album alb;
			//Artist art;
			//Genre gen;
			foreach(var m in new_media) {
				string key = lm.album_key(m);
				current_medias.set(m.rowid, 1);
				media_list.set((int)media_list.size(), m);
				
				if(update_album_view) {
					if((alb = album_keys.get(key)) == null) {
						alb = new Album(m.album_artist, m.album);
						album_keys.set(key, alb);
						album_list.set((int)album_list.size(), alb);
					}
					alb.add_media(m);
				}
				/*if(update_artist_view) {
					if((art =artist_keys.get(m.album_artist)) == null) {
						art = new Artist(m.album_artist);
						artist_keys.set(m.album_artist, art);
						artist_list.set((int)artist_list.size(), art);
					}
					art.add_media(m);
				}
				if(update_genre_view) {
					if((gen = genre_keys.get(m.genre)) == null) {
						gen = new Genre(m.genre);
						genre_keys.set(m.genre, gen);
						genre_list.set((int)genre_list.size(), gen);
					}
					gen.add_media(m);
				}*/
			}
			
			if(have_list_view)
				list_view.set_table(media_list);
			
			if(update_album_view && have_album_view)
				album_view.set_table(album_list);
			else
				album_needs_update = true;
				
			/*if(update_artist_view && have_artist_view)
				artist_view.set_table(artist_list);
			else
				artist_needs_update = true;
			
			if(update_genre_view && have_genre_view)
				genre_view.set_table(genre_list);
			else
				genre_needs_update = true;*/
			
			needs_update = false;
			in_update.unlock ();
			
			set_statusbar_info ();
			check_have_media ();
			update_library_window_widgets ();
			
		}
	}
	
	public void set_albums () {
		if(have_album_view) {
			in_update.lock ();
			
			var album_list = new HashTable<int, Album>(null, null);
			album_keys = new Gee.HashMap<string, Album>();
			
			Album alb;
			foreach(var m in list_view.get_table().get_values()) {
				string key = lm.album_key(m);
				
				if((alb = album_keys.get(key)) == null) {
					alb = new Album(m.album_artist, m.album);
					album_keys.set(key, alb);
					album_list.set((int)album_list.size(), alb);
				}
				alb.add_media(m);
			}
			
			album_view.set_table(album_list);
			album_needs_update = false;
			
			in_update.unlock ();
		}
	}
	
	void do_initial_populate() {
		needs_update = true;
		set_media(original_medias, true);
	}
}
